import pygame

"""
A gradient function is one that is given a number
between 0 and 1 and returns a color.

This module should provide some way to dynamically
generate gradient functions based on some parameters.
"""


def gradient_factory(color1, color2, exponent=1):
	"""Return a linear interpolation function between color1 and color2.

	>>> c1 = pygame.color.Color(0, 0, 0, 255)
	>>> c2 = pygame.color.Color(255, 255, 255, 255)
	>>> gradient = gradient_factory(c1, c2)
	>>> gradient(0)
	(0, 0, 0, 255)
	>>> gradient(1)
	(255, 255, 255, 255)
	>>> gradient(0.5)
	(127, 127, 127, 255)
	"""

	r = lambda x: int(color1.r * (1 - x) + color2.r * x)
	g = lambda x: int(color1.g * (1 - x) + color2.g * x)
	b = lambda x: int(color1.b * (1 - x) + color2.b * x)
	a = lambda x: int(color1.a * (1 - x) + color2.a * x)

	_gradient = lambda x: pygame.color.Color(r(x), g(x), b(x), a(x))
	gradient = lambda x: _gradient(x ** exponent)

	return gradient


def _interpolation_factory(*floats):
	"""Interpolate between an arbitrary number of numbers.

	Each argument may be a number or a tuple of two numbers. The second number in
	the tuple can be used to manipulate the distribution of the numbers in the
	interpolation function.

	>>> fn = _interpolation_factory((1, 2), 2, 4)
	>>> round(fn(0) * 10)
	11.0
	>>> round(fn(0.5) * 10)
	18.0
	>>> round(fn(0.667) * 10)
	21.0
	>>> round(fn(1) * 10)
	37.0
	"""

	widths = [i[1] if isinstance(i, tuple) else 1 for i in floats]
	floats = [i[0] if isinstance(i, tuple) else i for i in floats]

	centers = [0]
	for i in range(len(widths) - 1):
		centers.append(centers[-1] + (widths[i] + widths[i + 1]) / 2.0)
	full_width = centers[-1]

	def curve_factory(center, width):
		return lambda x: 1 / (1 + abs((float(x) - center) * 2 / width) ** 3)

	curves = [curve_factory(centers[i], widths[i]) for i in range(len(floats))]

	def interpolation(rate):
		factors = [c(rate * full_width) for c in curves]
		return sum([floats[i] * factors[i]
			for i in range(len(floats))]) / sum(factors)

	return interpolation


def _mix_factory(*colors):
	"""mix multiple colors."""

	colors = [i if isinstance(i, tuple) else (i, 1) for i in colors]

	r = _interpolation_factory(*[(c[0].r, c[1]) for c in colors])
	g = _interpolation_factory(*[(c[0].g, c[1]) for c in colors])
	b = _interpolation_factory(*[(c[0].b, c[1]) for c in colors])
	a = _interpolation_factory(*[(c[0].a, c[1]) for c in colors])

	return lambda x: pygame.color.Color(int(r(x)), int(g(x)), int(b(x)),
		int(a(x)))


class Background(object):
	"""Background gradient generator based on time of day.

	>>> gradient = Background.gradient(0)
	>>> gradient(0)
	(16, 17, 25, 255)
	"""

	backgrounds = [
		{  # night
			'top': pygame.color.Color(15, 15, 20, 255),
			'bottom': pygame.color.Color(30, 10, 80, 255),
			'exponent': 1,
			'duration': (6 - 0) * 2,
		},
		{  # morning
			'top': pygame.color.Color(60, 190, 220, 255),
			'bottom': pygame.color.Color(80, 230, 80, 255),
			'exponent': 1.6,
			'duration': 10 - 6,
		},
		{  # noon
			'top': pygame.color.Color(50, 30, 160, 255),
			'bottom': pygame.color.Color(60, 170, 200, 255),
			'exponent': 1,
			'duration': 17 - 10,
		},
		{  # dawn
			'top': pygame.color.Color(220, 80, 180, 255),
			'bottom': pygame.color.Color(50, 80, 180, 255),
			'exponent': 0.7,
			'duration': 19 - 17,
		},
		{  # twilight
			'top': pygame.color.Color(140, 30, 40, 255),
			'bottom': pygame.color.Color(30, 10, 100, 255),
			'exponent': 0.8,
			'duration': 22 - 19,
		},
		{  # night
			'top': pygame.color.Color(15, 15, 20, 255),
			'bottom': pygame.color.Color(30, 10, 80, 255),
			'exponent': 1,
			'duration': (24 - 22) * 2,
		},
	]

	color1 = staticmethod(_mix_factory(
		*[(bg['top'], bg['duration']) for bg in backgrounds]))
	color2 = staticmethod(_mix_factory(
		*[(bg['bottom'], bg['duration']) for bg in backgrounds]))
	exponent = staticmethod(_interpolation_factory(
		*[(bg['exponent'], bg['duration']) for bg in backgrounds]))

	@classmethod
	def gradient(cls, timeofday):
		return gradient_factory(cls.color1(timeofday), cls.color2(timeofday),
			exponent=cls.exponent(timeofday))


def draw(img, gradient):
	h = img.get_height()
	w = img.get_width()

	for y in range(h):
		color = gradient(y / float(h))
		pygame.draw.line(img, color, (0, y), (w - 1, y))



import math
from datetime import datetime

tilesize = 24  # pixel / tile
framerate = 60  # frames / second

gravity = 40.0
air_resistance = 0.01

"""
The speed is calculated by

	new_speed = old_speed * (1-air_resitance) + gravity

Therefore the maximum speed can be calculated by

	max_speed = gravity / air_resitance

We do not want the maximum speed to be more that tilesize
becausse then collision detection would not work.
So we will adjust the framerate to catch that case:
"""

if gravity / air_resistance > framerate ** 2:
	framerate = int(math.floor(math.sqrt(gravity / air_resistance)))


# We try to adopt to different tilesizes and framerates.
# However, rounding errors may have significant influence on game physics

gravity = gravity * tilesize / framerate ** 2


def timeofday():
	t = datetime.now().time()
	return (((t.microsecond / 1000000.0 + t.second) / 60.0 + t.minute) / 60.0
		+ t.hour) / 24.0 % 1



#!/usr/bin/env python

import pygame
from pygame.locals import QUIT, KEYDOWN, K_ESCAPE

from physics import tilesize, framerate
from level import LevelManager, Platform
from player import Player


def main():
	pygame.init()

	savegame = None

	sprites = pygame.sprite.OrderedUpdates()
	Player.groups = sprites,
	Platform.groups = sprites, LevelManager.platforms
	Player.collision_sprites = LevelManager.platforms

	level_manager = LevelManager()
	pygame.display.set_caption("Platformer Demo")
	screen = pygame.display.set_mode((level_manager.width * tilesize,
		level_manager.height * tilesize))

	# initiate objects
	player = Player()
	clock = pygame.time.Clock()

	while 1:
		clock.tick(framerate)
		sprites.update()

		# check for level change
		if player.rect.right > level_manager.width * tilesize:
			level_manager.right()
			player.rect.left = 0
		elif player.rect.left < 0:
			level_manager.left()
			player.rect.right = level_manager.width * tilesize
		elif player.rect.top < 0:
			level_manager.up()
			player.rect.bottom = level_manager.height * tilesize
		elif player.rect.bottom > level_manager.height * tilesize:
			level_manager.down()
			player.rect.top = 0

		for ev in pygame.event.get():
			if ev.type == QUIT:
				pygame.quit()
				return
			if ev.type == KEYDOWN:
				if ev.key == K_ESCAPE:
					pygame.quit()
					return
				elif ev.key == ord('s'):
					# save
					savegame = dict()
					savegame['pos'] = player.rect.x, player.rect.y
					savegame['level'] = level_manager.current
				elif ev.key == ord('l'):
					# load
					if savegame is None:
						raise Exception("no savegame available")
					else:
						level_manager.load(*savegame['level'])
						player.rect.x, player.rect.y = savegame['pos']
			player.handle_event(ev)

		# draw background
		screen.blit(level_manager.background, (0, 0))

		sprites.draw(screen)
		pygame.display.flip()


if __name__ == "__main__":
	main()



from random import random

import pygame

from physics import tilesize
from physics import timeofday
import gradient


class Platform(pygame.sprite.Sprite):
	"""One solid block."""
	def __init__(self, pos, color=(0, 0, 0)):
		super(Platform, self).__init__(self.groups)
		self.image = pygame.Surface((tilesize, tilesize))
		self.image.fill(color)
		self.rect = self.image.get_rect(topleft=pos)


class Level:
	"""A map containing platforms.

	A level typically fills the screen. If the player exits the level in any
	direction she enters the next one.
	"""

	def __init__(self, parent):
		self.parent = parent
		self.map = []
		for y in range(self.parent.height):
			line = []
			for x in range(self.parent.width):
				line.append('.')
			self.map.append(line)

	def load(self):
		"""Generate platforms from map."""
		for y in range(self.parent.height):
			for x in range(self.parent.width):
				if self.map[y][x] != ".":
					Platform((x * tilesize, y * tilesize))

	def get(self, x, y, default=None):
		try:
			return self.map[y][x]
		except IndexError:
			return default

	def to_int(self, x, y):
		a = self.get(x, y)
		if a is None:
			return 0
		elif a == '.':
			return -1
		else:
			return 1

	def __str__(self):
		return '\n'.join([''.join(row) for row in self.map])


class LevelManager:
	"""Generate and store all levels."""

	platforms = pygame.sprite.OrderedUpdates()

	def __init__(self):
		self._dict = dict()
		self.width = 24
		self.height = 12
		self.current = (0, 0)
		self.background = pygame.Surface((self.width * tilesize,
			self.height * tilesize))
		self.load(0, 0)

	def load(self, X, Y):
		"""Load level at position (X, Y)."""

		self.current = (X, Y)

		for sprite in self.platforms.sprites():
			sprite.kill()

		if (X, Y) not in self._dict:
			self.generate(X, Y)

		self._dict[(X, Y)].load()

		# also update gradient (only on level change for performance)
		gradient.draw(self.background, gradient.Background.gradient(timeofday()))

	def generate(self, X, Y):
		"""Generate a random level."""

		level = Level(self)

		x_min = 0
		x_max = self.width - 1
		y_min = 0
		y_max = self.height - 1

		# borders in adjacent levels should be the same
		if (X + 1, Y) in self._dict:
			x_max = self.width - 2
			for y in range(self.height):
				level.map[y][self.width - 1] = self._dict[(X + 1, Y)].map[y][0]

		if (X - 1, Y) in self._dict:
			x_min = 1
			for y in range(self.height):
				level.map[y][0] = self._dict[(X - 1, Y)].map[y][self.width - 1]

		if (X, Y + 1) in self._dict:
			y_max = self.height - 2
			for x in range(self.width):
				level.map[self.height - 1][x] = self._dict[(X, Y + 1)].map[0][x]

		if (X, Y - 1) in self._dict:
			y_min = 1
			for x in range(self.width):
				level.map[0][x] = self._dict[(X, Y - 1)].map[self.height - 1][x]

		def field(x, y, f=2, r=1.5, threshold=0):
				A = [
					level.to_int(x, y - 1),
					level.to_int(x, y + 1),
					level.to_int(x + 1, y),
					level.to_int(x - 1, y)]

				B = [
					level.to_int(x - 1, y - 1),
					level.to_int(x + 1, y - 1),
					level.to_int(x - 1, y + 1),
					level.to_int(x + 1, y + 1)]

				a = sum(A) * 1.0 / sum([abs(xx) for xx in A])
				b = sum(B) * 1.0 / sum([abs(xx) for xx in B])

				c = f * a - b
				d = 2 * r * random() - r

				return 'X' if c + d > threshold else '.'

		while x_min <= x_max and y_min <= y_max:
			for x in range(x_min, x_max + 1):
				level.map[y_min][x] = field(x, y_min)
			y_min += 1

			for y in range(y_min, y_max + 1):
				level.map[y][x_min] = field(x_min, y)
			x_min += 1

		self._dict[(X, Y)] = level

	# load() shortcuts
	def right(self):
		"""Load the level right of the current one."""
		X, Y = self.current
		self.load(X + 1, Y)

	def left(self):
		"""Load the level left of the current one."""
		X, Y = self.current
		self.load(X - 1, Y)

	def up(self):
		"""Load the level above the current one."""
		X, Y = self.current
		self.load(X, Y - 1)

	def down(self):
		"""Load the level below the current one."""
		X, Y = self.current
		self.load(X, Y + 1)



#!/usr/bin/env python

from setuptools import setup

setup(
    name='platformer-game',
    version='0.1.1',
    description="atmospheric platformer game",
    long_description=open('README.rst').read(),
    author='Tobias Bengfort',
    author_email='tobias.bengfort@gmx.net',
    packages=['platformer'],
    include_package_data=True,
    entry_points={'console_scripts': 'platformer=platformer:main'},
    install_requires=[
        'pygame',
    ],
    license='GPLv3+',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU General Public License v3 or later '
            '(GPLv3+)',
        'Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games',
    ])



import os
import math

import pygame
from pygame.locals import KEYDOWN, K_UP, K_DOWN, K_LEFT, K_RIGHT, K_SPACE
from pygame.locals import Rect

from physics import tilesize, framerate, air_resistance, gravity

LEFT = 0
RIGHT = 1

STAND = 0
WALK = 1
JUMP = 3
HANG = 4
CLIMB = 5


class Player(pygame.sprite.Sprite):
	def __init__(self):
		super(Player, self).__init__(self.groups)

		self.animation_duration = 0.1 * framerate
		self.jump_speed = 11.0 * tilesize / framerate
		self.speed = 3.0 * tilesize / framerate

		p = lambda s: os.path.join(os.path.dirname(__file__), s)
		self.img_default = pygame.image.load(p('sprites/heroine/default.png'))
		self.img_walk = pygame.image.load(p('sprites/heroine/walk.png'))
		self.img_fall = pygame.image.load(p('sprites/heroine/fall.png'))
		self.img_wall_1 = pygame.image.load(p('sprites/heroine/wall_1.png'))
		self.img_wall_2 = pygame.image.load(p('sprites/heroine/wall_2.png'))

		self.x_speed = 0
		self.y_speed = 0
		self.face = RIGHT
		self._animation_tick = 0
		self.set_animation(self.img_default)
		self.state = STAND

		# start position
		self.rect = self.image.get_rect(topleft=[5 * tilesize, 10 * tilesize])

	@property
	def image(self):
		return self.animation[0]

	def set_state(self, key):
		old_state = self.state

		if self.is_jumping():
			self.state = JUMP
		elif self.is_climbing():
			if key[K_UP]:
				self.state = CLIMB
			else:
				self.state = HANG
		else:
			if key[K_LEFT] or key[K_RIGHT]:
				self.state = WALK
			else:
				self.state = STAND

		return self.state != old_state

	def update(self):
		key = pygame.key.get_pressed()

		if self.is_climbing():
			if key[K_UP]:
				self.y_speed = -self.speed
			elif key[K_DOWN]:
				self.y_speed = self.speed + gravity * framerate / 20
			else:
				self.y_speed = gravity * framerate / 20

		if key[K_LEFT]:
			self.x_speed = -self.speed
			self.face = LEFT
		elif key[K_RIGHT]:
			self.x_speed = self.speed
			self.face = RIGHT
		else:
			self.x_speed = 0

		self.x_speed *= 1 - air_resistance
		self.y_speed *= 1 - air_resistance

		if self.is_jumping():
			self.y_speed += gravity

		if self.set_state(key):
			self.select_animation(self.state)
		self.do_animation()
		self.move(self.x_speed, self.y_speed)

	def set_animation(self, *animation):
		if self.face == RIGHT:
			self.animation = list(animation)
		else:
			flip = lambda i: pygame.transform.flip(i, True, False)
			self.animation = [flip(i) for i in animation]

	def do_animation(self):
		self._animation_tick += 1
		if self._animation_tick > self.animation_duration:
			self.animation.append(self.animation.pop(0))
			self._animation_tick = 0

	def select_animation(self, state):
		if state == STAND:
			self.set_animation(self.img_default)
		elif state == WALK:
			self.set_animation(self.img_default, self.img_walk)
		elif state == JUMP:
			self.set_animation(self.img_fall)
		elif state == HANG:
			self.set_animation(self.img_wall_1)
		elif state == CLIMB:
			self.set_animation(self.img_wall_1, self.img_wall_2)

	def is_climbing(self):
		if self.y_speed < -self.speed:
			return False

		new_rect = Rect(self.rect)
		new_rect.width += 2
		new_rect.x -= 1

		for sprite in self.collision_sprites:
			if new_rect.colliderect(sprite.rect):
				self.face = RIGHT if sprite.rect.x > new_rect.x else LEFT
				return True
		return False

	def is_jumping(self):
		if self.is_climbing():
			return False
		else:
			new_rect = Rect(self.rect)
			new_rect.y += 1

			for sprite in self.collision_sprites:
				if new_rect.colliderect(sprite.rect):
					return False
			return True

	def handle_event(self, ev):
		if ev.type == KEYDOWN:
			if ev.key == K_SPACE and not self.is_jumping():
				self.y_speed = -self.jump_speed

	def _move_x(self, dx):
		# Create a temporary new rect that has been moved by dx
		# and check for collisions

		new_rect = Rect(self.rect)
		new_rect.x += dx

		for sprite in self.collision_sprites:
			if new_rect.colliderect(sprite.rect):
				if dx > 0:  # moving right
					return sprite.rect.left - self.rect.right
				else:  # moving left
					return sprite.rect.right - self.rect.left

		return dx

	def _move_y(self, dy):
		# Create a temporary new rect that has been moved by dy
		# and check for collisions

		new_rect = Rect(self.rect)
		new_rect.y += dy

		for sprite in self.collision_sprites:
			if new_rect.colliderect(sprite.rect):
				if dy > 0:  # moving down
					# Landed!
					self.y_speed = 0
					return sprite.rect.top - self.rect.bottom
				elif dy < 0:  # moving up
					# oww, we hit our head
					self.y_speed = 0
					return sprite.rect.bottom - self.rect.top

		return dy

	def move(self, dx, dy):
		self.rect.x += self._move_x(_ceil(dx))
		self.rect.y += self._move_y(_ceil(dy))


def _ceil(x):
	"""
	>>> _ceil(1.5)
	2.0
	>>> _ceil(-1.5)
	-2.0
	"""
	return math.ceil(x) if x > 0 else math.floor(x)
